/* sslfilter.c */
/* Copyright (C) 1995 Eric Young (eay@mincom.oz.au).
 * All rights reserved.
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * See the COPYRIGHT file in the SSLeay distribution for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h> /* Needed under linux for FD_XXX */
#include <netinet/in.h>
#include "net.h"
#include "X509.h"
#include "ssl.h"
#include "ssl_apps.h"

/*#define TEST_CERT "client.PEM" */ /* no default cert. */

#define BUFSIZZ	40*1024

int verify_depth=0;
int verify_error=VERIFY_OK;

#define SSL_TO_CLEAR	1
#define CLEAR_TO_SSL	2
#define SSL_TO_SSL	3
#define CLEAR_TO_CLEAR	4

typedef struct connection_st
	{
	int filter;
	int fork;

	char *CAfile;	/* a CA file */
	char *CApath;	/* the CA path to CA's */
	int verify;	/* verify mode */
	char *cipher;	/* prefered cipher */
	char *cert_file;/* certificate to use */
	char *key_file; /* private key file */

	SSL *ssl[2];       /* SSL structure for reading/writing */
	char *host[2];	/* who has connected. */
	int port[2];	/* local port to listen too. */
	int socket[2];	/* socket */
	int readfd[2];   /* fd for normal read */
	int writefd[2];  /* fd for normal write */
	} CON;

#ifdef PROTO
static int setup_connections (CON *c, int argc, char **argv);
static int process (CON *c);
static int get_connection(CON *con, int port, int *ret);
#else
static int setup_connections ();
static int process ();
static int get_connection();
#endif

void usage()
	{
	fprintf(stderr,"usage: client args\n");
	fprintf(stderr,"\n");
	}

void zero_con(con)
CON *con;
	{
	int i;

	con->CAfile=	(char *)X509_get_default_cert_file();
	con->CApath=	(char *)X509_get_default_cert_dir();
	con->verify=	SSL_VERIFY_NONE;
	con->cipher=	NULL;
	con->cert_file=	NULL;
	con->key_file=	NULL;
	con->filter=	SSL_TO_CLEAR; /* con1 to con2 */
	con->fork=	0;

	for (i=0; i<2; i++)
		{
		con->ssl[i]=	(SSL *)SSL_new();
		con->host[i]=	NULL;
		con->port[i]=	-1;
		con->socket[i]=	-1;
		con->readfd[i]=	-1;
		con->writefd[i]=-1;
		}
	}

CON *glob_con;
int accept_sock;

main(argc, argv)
int argc;
char **argv;
	{
	CON con;

	zero_con(&con);

	argc--;
	argv++;
	while (argc >= 1)
		{
		if ((*argv)[0] != '-') break;
		if	(strcmp(*argv,"-host1") == 0)
			{
			if (argc-- < 1) goto bad;
			con.host[0]= *(++argv);
			}
		else if	(strcmp(*argv,"-host2") == 0)
			{
			if (argc-- < 1) goto bad;
			con.host[1]= *(++argv);
			}
		else if	(strcmp(*argv,"-port1") == 0)
			{
			if (argc-- < 1) goto bad;
			con.port[0]=atoi(*(++argv));
			if (con.port[0] == 0) goto bad;
			}
		else if	(strcmp(*argv,"-port2") == 0)
			{
			if (argc-- < 1) goto bad;
			con.port[1]=atoi(*(++argv));
			if (con.port[1] == 0) goto bad;
			}

		else if	(strcmp(*argv,"-verify") == 0)
			{
			con.verify=SSL_VERIFY_PEER;
			if (argc-- < 1) goto bad;
			verify_depth=atoi(*(++argv));
			}
		else if	(strcmp(*argv,"-Verify") == 0)
			{
			con.verify=SSL_VERIFY_PEER|
				SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
			if (argc-- < 1) goto bad;
			verify_depth=atoi(*(++argv));
			}
		else if	(strcmp(*argv,"-cert") == 0)
			{
			if (argc-- < 1) goto bad;
			con.cert_file= *(++argv);
			}
		else if	(strcmp(*argv,"-key") == 0)
			{
			if (argc-- < 1) goto bad;
			con.key_file= *(++argv);
			}
		else if	(strcmp(*argv,"-CApath") == 0)
			{
			if (argc-- < 1) goto bad;
			con.CApath= *(++argv);
			}
		else if	(strcmp(*argv,"-CAfile") == 0)
			{
			if (argc-- < 1) goto bad;
			con.CAfile= *(++argv);
			}
		else if	(strcmp(*argv,"-cipher") == 0)
			{
			if (argc-- < 1) goto bad;
			con.cipher= *(++argv);
			}
		else if (strcmp(*argv,"-ssl2clear") == 0)
			con.filter=SSL_TO_CLEAR;
		else if (strcmp(*argv,"-clear2ssl") == 0)
			con.filter=CLEAR_TO_SSL;
		else if (strcmp(*argv,"-clear2clear") == 0)
			con.filter=CLEAR_TO_CLEAR;
		else if (strcmp(*argv,"-ssl2ssl") == 0)
			con.filter=SSL_TO_SSL;
		else
			{
			fprintf(stderr,"unknown option %s\n",*argv);
			goto bad;
			}
		argc--;
		argv++;
		}

	if ((!X509_load_verify_locations(con.CAfile,con.CApath)) ||
		(!X509_set_default_verify_paths()))
		{
		fprintf(stderr,"X509_load_verify_locations error:%s\n",
			X509_error_string(X509_errno));
		exit(1);
		}

	setup_connections(&con,argc,argv);

	if (con.ssl[0] != NULL)
		SSL_set_verify(con.ssl[0],con.verify,verify_callback);
	if (con.ssl[0] != NULL)
		SSL_set_verify(con.ssl[1],con.verify,verify_callback);

	
	if (con.cipher == NULL)
		{
		if (con.ssl[0] != NULL)
			SSL_set_pref_cipher(con.ssl[0],getenv("SSL_CIPHER"));
		if (con.ssl[1] != NULL)
			SSL_set_pref_cipher(con.ssl[1],getenv("SSL_CIPHER"));
		}
	else
		{
		if (con.ssl[0] != NULL)
			SSL_set_pref_cipher(con.ssl[0],con.cipher);
		if (con.ssl[1] != NULL)
			SSL_set_pref_cipher(con.ssl[1],con.cipher);
		}

	if (con.ssl[0] != NULL)
		set_cert_stuff(con.ssl[0],con.cert_file,con.key_file);
	if (con.ssl[1] != NULL)
		set_cert_stuff(con.ssl[1],con.cert_file,con.key_file);

	argv++;
	argc--;

	process(&con);
	/* never return */

	exit(0);
bad:
	usage();
	exit(1);
	}

static int setup_connections(c, argc, argv)
CON *c;
int argc;
char **argv;
	{
	int zero,one,i;

	if ((c->port[0] == -1) && (c->host[0] != NULL))
		{
		fprintf(stderr,"you have not specified a port for host1\n");
		exit(1);
		}

	if ((c->port[1] == -1) && (c->host[1] != NULL))
		{
		fprintf(stderr,"you have not specified a port for host[1]\n");
		exit(1);
		}

	/* stdio(1) <convert> program(2) */
	if ((c->port[0] == -1) && (c->port[1] == -1))
		{
		c->socket[0]=-1;
		c->readfd[0]=fileno(stdin);
		c->writefd[0]=fileno(stdout);
		c->socket[1]=-1;
		if (argc > 0)
			{
			i=spawn(argc,argv,&(c->readfd[1]),&(c->writefd[1]));
			if (i <= 0)
				{
				perror("spawn");
				exit(1);
				}
			}
		}
		/* program <convert> host[1] */
	else if ((c->port[0] == -1) || (c->port[1] == -1))
		{
		if (c->port[0] == -1)
			{ zero=0; one=1; }
		else	{ zero=1; one=0; }
		if (c->host[one] != NULL)
			{
			if (!init_client(&(c->socket[one]),
				c->host[one],c->port[one]))
				{
				perror("init_client");
				exit(1);
				}
			}
		else
			{
			if (!get_connection(c,c->port[one],&(c->socket[one])))
				{
				perror("gen_connection");
				exit(1);
				}
			}
		c->readfd[one]=c->socket[one];
		c->writefd[one]=c->socket[one];
		c->socket[zero]= -1;
		if (argc > 0)
			i=spawn(argc,argv,&(c->readfd[zero]),&(c->writefd[zero]));
		else
			{
			c->readfd[zero]=fileno(stdin);
			c->writefd[zero]=fileno(stdout);
			}
		}
	else /* ip:port <conv> ip:port */
		{
		for (i=0; i<2; i++)
			{
			if (c->host[i] != NULL)
				{
				if (!init_client(&(c->socket[i]),
					c->host[i],c->port[i]))
					{
					perror("init_client");
					exit(1);
					}
				}
			else
				{
				if (!get_connection(c,c->port[i],
					&(c->socket[i])))
					{
					perror("gen_connection");
					exit(1);
					}
				}
			c->readfd[i]=c->socket[i];
			c->writefd[i]=c->socket[i];
			}
		}
	/* we now have connections setup. */	

	if (((c->filter == SSL_TO_SSL) && 
		((c->socket[0] == -1) || c->socket[1] == -1)) ||
		((c->filter == SSL_TO_CLEAR) && (c->socket[0] == -1)) ||
		((c->filter == CLEAR_TO_SSL) && (c->socket[1] == -1)))
		{
		fprintf(stderr,"SSL will only be run over a socket\n");
		exit(1);
		}
/* NO SSL */
	c->ssl[0]=c->ssl[1]=NULL;
	}

static int process(c)
CON *c;
	{
	int width,w,i,j,skip,tot,n;
	fd_set readfds,oreadfds;
	char buf[BUFSIZZ];

	width=0;
	FD_ZERO(&readfds);
	for (i=0; i<2; i++)
		{
		FD_SET(c->readfd[i],&readfds);
		if (c->readfd[i] > width) width=c->readfd[i];
		}
	width++;
	for (;;)
		{
		n=0;
		w=0;
		skip=0;
		for (n=0; n<2; n++)
			{
			if ((c->ssl[n] != NULL) && SSL_pending(c->ssl[n]))
				skip=1;
			}

		if (!skip)
			{
			memcpy(&oreadfds,&readfds,sizeof(readfds));
			i=select(width,&oreadfds,NULL,NULL,NULL);
			if (i < 0) break;
fprintf(stderr,"%d\n",i);
			}
		for (n=0; n < 2; n++)
			{
			w=0;
			if (!skip && (c->ssl[n] == NULL))
				{
				if (FD_ISSET(c->readfd[n],&oreadfds))
					{
					i=read(c->readfd[n],buf,BUFSIZZ);
fprintf(stderr,"r=%d\n",i);
					if (i <= 0) goto done; /**/
					w++;
					}
				}
			else /* SSL_read */
				{
				if (SSL_pending(c->ssl[n]) ||
					(!skip &&
					  FD_ISSET(c->readfd[n],&oreadfds)))
					{
					i=SSL_read(c->ssl[n],buf,BUFSIZZ);
fprintf(stderr,"R=%d\n",i);
					if (i <= 0) goto done; /**/
					w++;
					}
				}
			if (w)
				{
				if (c->ssl[!n] != NULL)
					{
					if (SSL_write(c->ssl[!n],buf,i) <= 0) goto done;
					}
				else
					{
					tot=0;

					for (;;)
						{
						j=write(c->writefd[!n],&(buf[tot]),i);
						if (j <= 0) goto done;
						if (tot+j >= i) break;
						i-=j;
						tot+=j;
						}
					}
				}
			}
		}
done:
	if (c->socket[0] != -1) close(c->socket[0]);
	if (c->socket[1] != -1) close(c->socket[1]);
	return(1);
	}

static int get_connection(con, port, ret)
CON *con;
int port;
int *ret;
	{
	int sock;
	char *name;
	int accept_socket;

	if (!init_server(&accept_socket,port)) exit(1);

	for (;;)
		{
		if (do_accept(accept_socket,&sock,&name) == 0)
			exit(1);
		if (con->fork)
			{
			if (fork() == 0) break;
			free(name);
			close(sock);
			continue;
			}
		break;
		}
	*ret=sock;
	}

